use crate::chacha_io::{ChaChaReader, ChaChaWriter};
use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
use chrono::*;

use rand::Rng;
use sha2::{Digest, Sha256};

use std::convert::TryInto;

use std::fs::File;
use std::io;
use std::io::prelude::*;
use std::iter::Iterator;
use std::path::{Path, PathBuf};

pub struct Snippet<'a> {
    pub id: String,
    pub storage: &'a SnippetStorage<'a>,
}

#[allow(dead_code)]
pub struct SnippetMeta {
    created: DateTime<Utc>,
    compression: Option<String>,
}

pub struct SnippetStorage<'a> {
    directory: &'a Path,
}

impl<'a> Snippet<'a> {
    pub fn random(storage: &'a SnippetStorage) -> Snippet<'a> {
        Snippet::new(
            &rand::thread_rng()
                .gen_ascii_chars()
                .take(8)
                .collect::<String>(),
            storage,
        )
    }

    pub fn new(id: &str, storage: &'a SnippetStorage) -> Snippet<'a> {
        Snippet {
            id: id.to_string(),
            storage,
        }
    }

    pub fn file_id(&self) -> String {
        SnippetStorage::file_id(&self.id)
    }

    pub fn passphrase(&self) -> ([u8; 8], [u8; 32]) {
        let mut hasher = Sha256::new();
        hasher.input(self.id.as_bytes());
        let res = hasher.result();
        let nonce: [u8; 8] = res[0..8].try_into().unwrap();
        let mut hasher = Sha256::new();
        hasher.input(self.id.as_bytes());
        hasher.input(b"pass");
        let pass: [u8; 32] = res[0..32].try_into().unwrap();
        (nonce, pass)
    }

    pub fn path(&self) -> PathBuf {
        self.storage.directory.join(&self.file_id())
    }

    pub fn metadata(&self) -> Result<SnippetMeta, io::Error> {
        let mut file = File::open(self.path())?;
        let (nonce, key) = self.passphrase();
        self.metadata_via_handle(&mut ChaChaReader::new(&key, &nonce, &mut file))
    }

    fn metadata_via_handle(&self, hdl: &mut impl Read) -> Result<SnippetMeta, io::Error> {
        let timestamp = hdl.read_i64::<BigEndian>()?;
        let comp_len = hdl.read_u16::<BigEndian>()? as usize;
        let mut comp = Vec::with_capacity(comp_len);
        comp.resize(comp_len, 0u8);
        hdl.read_exact(&mut comp)?;
        let comp = String::from_utf8(comp).unwrap();
        Ok(SnippetMeta {
            created: Utc.timestamp(timestamp, 0),
            compression: Some(comp).filter(|_| comp_len > 0),
        })
    }

    pub fn contents(&self) -> Result<String, io::Error> {
        let mut file = File::open(self.path())?;
        let (nonce, key) = self.passphrase();
        let mut reader = ChaChaReader::new(&key, &nonce, &mut file);
        let meta = self.metadata_via_handle(&mut reader)?;
        fn read_string(r: &mut impl Read) -> Result<String, io::Error> {
            let mut text = String::new();
            r.read_to_string(&mut text)?;
            Ok(text)
        }
        dbg!((&meta.compression, &meta.created, self.file_id()));
        match meta.compression {
            Some(ref comp) if comp == "snap" => {
                let mut r = snap::Reader::new(&mut reader);
                read_string(&mut r)
            }
            _ => read_string(&mut reader),
        }
    }

    pub fn write(self, content: &str) -> Result<Snippet<'a>, io::Error> {
        let mut file = File::create(self.path())?;
        let (nonce, key) = self.passphrase();
        let mut writer = ChaChaWriter::new(&key, &nonce, &mut file);
        writer.write_i64::<BigEndian>(Utc::now().timestamp())?;
        let comp = if content.len() > 2048 {
            Some("snap")
        } else {
            None
        };
        writer.write_u16::<BigEndian>(comp.map(|s| s.len() as u16).unwrap_or(0u16))?;
        writer.write(comp.map(|s| s.as_bytes()).unwrap_or(&[0u8; 0]))?;
        match comp {
            Some(ref comp) if comp == &"snap" => {
                let mut w = snap::Writer::new(&mut writer);
                w.write_all(content.as_bytes())?
            }
            _ => writer.write_all(content.as_bytes())?,
        };
        Ok(Snippet::new(&self.id, self.storage))
    }
}

impl<'a> SnippetStorage<'a> {
    pub fn new(directory: &'a Path) -> SnippetStorage<'a> {
        SnippetStorage {
            directory: directory,
        }
    }
    pub fn file_id(id: &str) -> String {
        let mut hasher = Sha256::new();
        hasher.input(id.as_bytes());
        hex::encode(&hasher.result()[0..12])
    }
    fn has(&self, id: &str) -> bool {
        self.directory.join(Self::file_id(id)).exists()
    }

    pub fn open(&self, id: &str) -> Option<Snippet> {
        if self.has(id) {
            Some(Snippet::new(id, self))
        } else {
            None
        }
    }
}
